use std::time::Instant;

use crossterm::event;
use tui::{
	backend::Backend,
	layout::{Alignment, Constraint, Direction, Layout, Margin},
	style::{Color, Modifier, Style},
	text::Span,
	widgets::{Block, Borders, List, ListItem, ListState, Paragraph},
	Terminal,
};
use tui_textarea::{Input, Key, TextArea};

use crate::index::IndexConnection;

enum SearchMode {
	Simple,
	Advanced,
}

impl SearchMode {
	fn prompt(&self) -> Span {
		match self {
			SearchMode::Simple => Span {
				content: "[simple] ".into(),
				style: Style {
					fg: Some(Color::LightCyan),
					..Default::default()
				},
			},
			SearchMode::Advanced => Span {
				content: "[advanced] ".into(),
				style: Style {
					fg: Some(Color::LightRed),
					..Default::default()
				},
			},
		}
	}
}

pub fn run<B: Backend>(index_path: &str, term: &mut Terminal<B>) -> anyhow::Result<()> {
	let index = IndexConnection::open(index_path)?;
	let mut mode = SearchMode::Simple;
	let mut textarea = TextArea::new(vec![]);
	let mut results = Ok(Vec::new());
	let mut elapsed = 0f32;
	let mut more_results = "";
	let mut query = String::new();

	let cursor_style = textarea.cursor_style();
	let cursor_line_style = textarea.cursor_line_style();

	let mut list_focused = false;
	let mut list_state = ListState::default();

	loop {
		if list_focused {
			if let (Some(selected), Ok(results)) = (list_state.selected(), &mut results) {
				if selected == results.len() - 1 {
					let before = Instant::now();
					if !query.is_empty() {
						// if the same query already succeeded, another one shouldn't fail, so we can use the ? operator
						let mut new_results = index.search(&query, 100, results.len() as u64)?;
						more_results = if new_results.len() == 100 { "+" } else { "" };
						results.append(&mut new_results);
					}
					elapsed = before.elapsed().as_secs_f32() * 1000.0;
				}
			}
		} else {
			let new_query = textarea.lines()[0].as_str();
			if new_query.is_empty() {
				results = Ok(Vec::new());
				more_results = "";
			} else if query != new_query {
				let before = Instant::now();
				query = new_query.to_string();
				list_state.select(None);
				results = index.search(new_query, 100, 0);
				if let Ok(results_len) = results.as_ref().map(|r| r.len()) {
					more_results = if results_len == 100 { "+" } else { "" };
				}
				elapsed = before.elapsed().as_secs_f32() * 1000.0;
			}
		}
		term.draw(|f| {
			let chunks = Layout::default()
				.direction(Direction::Vertical)
				.constraints([
					Constraint::Length(1),
					Constraint::Min(1),
					Constraint::Length(1),
				])
				.split(f.size());

			let mode_prompt = mode.prompt();
			let input_chunks = Layout::default()
				.direction(Direction::Horizontal)
				.constraints([
					Constraint::Length(mode_prompt.width() as u16),
					Constraint::Min(2),
				])
				.split(chunks[0]);

			let prompt_paragraph = Paragraph::new(mode_prompt);
			f.render_widget(prompt_paragraph, input_chunks[0]);
			f.render_widget(textarea.widget(), input_chunks[1]);

			let list_chunk = chunks[1].inner(&Margin {
				vertical: 0,
				horizontal: 1,
			});

			let (messages, title) = match &results {
				Ok(msgs) => {
					let termwidth = if list_chunk.width > 10 {
						list_chunk.width - 3
					} else {
						list_chunk.width
					} as usize;
					let msgs: Vec<ListItem> = msgs
						.iter()
						.map(|msg| {
							msg.listitem(termwidth).unwrap_or_else(|e| {
								ListItem::new(Span::styled(
									format!("Error: {e}"),
									Style::default().fg(Color::LightRed),
								))
							})
						})
						.collect();
					let title = Span::raw(format!(
						"{}{} results in {:.3}ms ",
						msgs.len(),
						more_results,
						elapsed
					));
					(msgs, title)
				}
				Err(e) => (
					Vec::new(),
					Span::styled(e.to_string(), Style::default().fg(Color::LightRed)),
				),
			};

			let block = Block::default()
				.title(title)
				.title_alignment(Alignment::Right)
				.borders(Borders::TOP | Borders::BOTTOM)
				.border_style(Style::default().fg(Color::Green));

			let list = List::new(messages)
				.block(block)
				.style(Style::default().fg(Color::White))
				.highlight_style(
					Style::default()
						.add_modifier(Modifier::BOLD)
						.add_modifier(Modifier::ITALIC),
				)
				.highlight_symbol("> ");
			f.render_stateful_widget(list, list_chunk, &mut list_state);

			let keybinds_hint = Paragraph::new("Quit: C-q  Switch focus: C-Tab  Settings: C-k")
				.style(Style::default().fg(Color::Green));
			f.render_widget(keybinds_hint, chunks[2]);
		})?;

		let event = Input::from(event::read()?);

		match event {
			Input {
				key: Key::Char('q'),
				ctrl: true,
				..
			} => break,
			Input {
				key: Key::Char('k'),
				ctrl: true,
				..
			} => break,
			_ => {}
		}

		if list_focused {
			match event {
				Input {
					key: Key::Up,
					ctrl: true,
					..
				} => {
					list_focused = false;
					list_state.select(None);
					textarea.set_cursor_style(cursor_style);
					textarea.set_cursor_line_style(cursor_line_style);
				}
				Input {
					key: Key::Down,
					ctrl: false,
					..
				} => {
					if let Some(selected) = list_state.selected() {
						if let Ok(results) = &results {
							if selected < results.len() - 1 {
								list_state.select(Some(selected + 1));
							}
						}
					}
				}
				Input {
					key: Key::Up,
					ctrl: false,
					..
				} => {
					if let Some(selected) = list_state.selected() {
						if selected > 0 {
							list_state.select(Some(selected - 1));
						}
					}
				}
				Input {
					key: Key::Enter, ..
				} => {}
				_ => {}
			}
		} else {
			match event {
				Input {
					key: Key::Down,
					ctrl: true,
					..
				}
				| Input {
					key: Key::Enter, ..
				} => {
					list_focused = true;
					list_state.select(Some(0));
					textarea.set_cursor_line_style(Style::default());
					textarea.set_cursor_style(Style::default());
				}
				Input {
					key: Key::Char('m'),
					ctrl: true,
					..
				} => {}
				Input {
					key: Key::Char('m'),
					alt: true,
					..
				} => match mode {
					SearchMode::Simple => mode = SearchMode::Advanced,
					_ => mode = SearchMode::Simple,
				},
				input => {
					textarea.input(input);
				}
			}
		}
	}
	Ok(())
}
