use rusqlite::Connection;

use crate::dce;

mod models;

pub struct IndexConnection(Connection);

impl IndexConnection {
	pub fn open(path: &str) -> rusqlite::Result<Self> {
		let db = Connection::open(path)?;
		db.execute_batch(include_str!("schema.sql"))?;
		Ok(Self(db))
	}

	pub fn add_channel(&self, channel: dce::Channel) -> rusqlite::Result<usize> {
		let mut stmt = self
			.0
			.prepare_cached("insert or replace into channels(id,name) values (?,?)")?;
		stmt.execute((channel.id, channel.name))
	}

	pub fn add_msg(&self, msg: dce::Message, channel: u64) -> rusqlite::Result<usize> {
		let mut stmt = self.0.prepare_cached("insert or replace into msgs(id,body,author,attachments,timestamp,channel) values (?,?,?,?,?,?)")?;
		let attachments = msg
			.attachments
			.into_iter()
			.map(|a| a.url)
			.collect::<Vec<_>>();
		let attachments = if attachments.is_empty() {
			None
		} else {
			Some(attachments.join("; "))
		};
		stmt.execute((
			msg.id,
			msg.content,
			format!("{}#{}", msg.author.name, msg.author.discriminator),
			attachments,
			msg.timestamp.timestamp(),
			channel,
		))
	}

	pub fn search(
		&self,
		q: &str,
		limit: u32,
		offset: u64,
	) -> rusqlite::Result<Vec<models::Message>> {
        let mut stmt = self.0.prepare_cached("select m.id,m.timestamp,m.body,m.author,m.attachments,c.name from msgs_fts as f inner join msgs as m on f.rowid=m.id inner join channels as c on m.channel=c.id where f.body match ? order by rank limit ? offset ?")?;
		let rows = stmt.query_map((q, limit, offset), |r| {
			Ok(models::Message {
				id: r.get(0)?,
				timestamp: r.get(1)?,
				content: r.get(2)?,
				author: r.get(3)?,
				attachments: r.get(4)?,
				channel: r.get(5)?,
			})
		})?;
		let mut messages = Vec::new();
		for msg in rows {
			messages.push(msg?);
		}
		Ok(messages)
	}

	pub fn optimize(&self) -> rusqlite::Result<()> {
		self.0
			.execute("insert into msgs_fts(msgs_fts) values('optimize')", [])?;
		self.0.execute("vacuum", [])?;
		Ok(())
	}
}
