use anyhow::anyhow;
use chrono::{offset::TimeZone, Local, NaiveDateTime};
use tui::{
	style::{Color, Style},
	text::{Span, Spans},
	widgets::ListItem,
};

use crate::strchunks::StrChunks;

pub struct Message {
	pub id: i64,
	pub timestamp: i64,
	pub content: String,
	pub author: String,
	pub attachments: Option<String>,
	pub channel: String,
}

impl Message {
	pub fn listitem(&self, termsize: usize) -> anyhow::Result<ListItem> {
		let datetime = Local
			.from_utc_datetime(
				&NaiveDateTime::from_timestamp_opt(self.timestamp, 0)
					.ok_or_else(|| anyhow!("Failed to parse timestamp {}", self.timestamp))?,
			)
			.format("%F %T");

		let header = vec![
			Span::styled("#", Style::default().fg(Color::LightBlue)),
			Span::styled(&self.channel, Style::default().fg(Color::LightBlue)),
			Span::raw(" "),
			Span::styled(&self.author, Style::default().fg(Color::LightGreen)),
			Span::styled(format!(" {datetime}"), Style::default().fg(Color::DarkGray)),
		];

		let mut spans = vec![Spans(header)];
		spans.extend(StrChunks::new(&self.content, termsize).map(Spans::from));

		if let Some(attachments) = &self.attachments {
			spans.extend(
				StrChunks::new(&format!("Attachments: {}", attachments), termsize).map(|v| {
					Spans::from(Span::styled(
						v.to_string(),
						Style::default().fg(Color::DarkGray),
					))
				}),
			);
		}

		spans.push(Spans::from(""));

		Ok(ListItem::new(spans))
	}
}
